#ifndef DICTIONARY_H
#define DICTIONARY_H

/*
 * REFLEX - Real-time Event FLow EXecutive
 *
 * A lightweight operating system for deeply embedded systems.
 *
 *
 * Class(ses): Dictionary
 * Revision: $Id:$
 * Author: Steffen Schulze
 *
 * Description: Dictionary Datatype realised in flat memory
 *
 * (c) Steffen Schulze, BTU Cottbus 2005
 */
#include "../config.h"

template<typename T> class Vector2D
{
	public:
		
		Vector2D()
		{
		}
		
		Vector2D(T _a, T _b) : a(_a), b(_b)
		{
		}
		
		T	a;
		T	b;
};


/**
 * This class implements a simple dictionary without dynamic
 * memory. It uses only an array as classmember variable, which
 * stores the entries.
 *
 * @author Steffen Schulze
 * @version $Id$
 */
template <const UINT16 DIC_SIZE> 
class Dictionary
{
	public:
		Dictionary():currIndex(0), currPos(DIC_SIZE - 1), maxLength(0)
		{
		}
		
		~Dictionary()
		{
		}
		
		void initForLZW();
		void reset();
		
		/**
		 * Adds a String into the Dictionary
		 *
		 * @param pString the pointer to the string
		 * @param length is the lenght of the string
		 * @return @c >0 if adding the string was succesful, @c 0 if the dictionary is full
		 */
		UINT16 addString(BYTE* pString, UINT16 length);

		UINT16 getLastIndex()
		{
			return currIndex;
		}
				
		UINT16 getSize()
		{
			return DIC_SIZE - currPos;
		}
		
		UINT16 getMaxLength()
		{
			return maxLength;
		}
		
		/**
		 * Search the given string in the dictionary
		 *
		 * @param pString the pointer to the string
		 * @param length is the lenght of the string
		 * @return @c 0 if the string is not in dictionart, @c index of the string
		 */
		//UINT16 searchString(BYTE* pString, UINT16 length);
	
		/**
		* Search longest phrase of given string in dictionary.
		*
		* @param pString the pointer to the string
		* @param length is the lenght of the string
		* @return @c 0 if no phrase is in dictionart, @c index of the longest phrase and the length of the phrase
		*/
		Vector2D<UINT16> searchPhrase(BYTE* pString, UINT16 length);
		
	
		/**
		* Get String in the given index of the dictionary
		*
		* @param index the entry in the dictionary
		* @param pString the pointer to the string
		* @return @c 0 if the index is not in dictionary, @c length of returned string
		*/
		UINT16	getString(UINT16 index, BYTE* pString);
		
		#ifdef TEXTDEBUG
		UINT16 printAll(UINT16 hideIndex = 0)
		{
			UINT16 searchIndex	= currIndex;
			UINT16 searchPos	= currPos + 1;
			
			fprintf(stderr, "\t\t\t\tprintALL\n");
			
			//for each entry
			while(hideIndex < searchIndex--)
			{
				
				//read length of string in current entry
				UINT16 stringLength = dicArray[searchPos++] << 8;
				stringLength 		+=  dicArray[searchPos++];
				
				fprintf(stderr, "\t\t\t\t\t%d : %d :", searchIndex + 1, stringLength);
				
				//compare searched string and string in entry
				UINT16 stringPos	= 0;
				while(stringPos < stringLength)
				{
					fprintf(stderr, "%c", dicArray[searchPos++]);
					stringPos++;
				}
				
				fprintf(stderr, "\n");
			}
			
			return 0;
		}
		#endif
	
		
	protected:
		
		enum
		{
			/*
			 * Constant headersize of entries
			 */
			DIC_HEADER_SIZE = 2
		};
		
		/*
		 * Stores current index
		 */
		UINT16 currIndex;
		
		/*
		 * Stores current position in dicArray
		 */
		UINT16 currPos;
		
		/*
		 * Length of longest String in dictionary
		 */
		UINT16 maxLength;
		
		//Memory for the dictionary
		BYTE dicArray[DIC_SIZE];
};



////////////////////////////////////////////

template <const UINT16 DIC_SIZE> 
		void Dictionary<DIC_SIZE>::initForLZW()
{
	#ifdef TEXTDEBUG
	fprintf(stderr, "\t\t\t\t\t> initForLZW\n");
	#endif

	//init the dictionary
	BYTE string[1];
	for(UINT16 i = 0; i < 256; i++)
	{
		string[0] = i;
		addString(string,1);
	}

	#ifdef TEXTDEBUG
	fprintf(stderr, "\t\t\t\t\t< initForLZW\n");
	#endif
	
}





template <const UINT16 DIC_SIZE> 
void Dictionary<DIC_SIZE>::reset()
{
	#ifdef TEXTDEBUG
	fprintf(stderr, "\t\t\t\t\t> reset\n");
	#endif

	currPos = DIC_SIZE - 1;
	currIndex = 0;
	maxLength = 0;
	
	#ifdef TEXTDEBUG
	fprintf(stderr, "\t\t\t\t\t< reset\n");
	#endif
	
}



template <const UINT16 DIC_SIZE> 
UINT16 Dictionary<DIC_SIZE>::addString(BYTE* pString, UINT16 length)
{
	if(currPos < (length + DIC_HEADER_SIZE))
	{
		//dictionary is full
		return 0;
	}
	
	if(length > maxLength)
	{
		maxLength = length;
	}
	
	UINT16 stringPos = length;
	while(stringPos--)
	{
		dicArray[currPos--] = pString[stringPos];
	}
	
	
	//writes the 16bit length value into the next 2 bytes
	dicArray[currPos--] = length & 0xFF;
	dicArray[currPos--] = length >> 8;
	
	currIndex++;
	return currIndex;
}
/*
template <const UINT16 DIC_SIZE> 
UINT16 Dictionary<DIC_SIZE>::searchString(BYTE* pString, UINT16 length)
{
	UINT16 searchIndex	= currIndex;
	UINT16 searchPos	= currPos + 1;
	
	//for each entry
	while(searchIndex--)
	{
		
		//read length of string in current entry
		UINT16 stringLength = dicArray[searchPos++] << 8;
		stringLength 		+=  dicArray[searchPos++];
		
		//if length of searched string and current entry is uneven, then go to next entry
		if(length != stringLength)
		{
			searchPos += stringLength;
			continue;
		}
		
		//compare searched string and string in entry
		UINT16 stringPos = 0;
		while(stringPos < stringLength)
		{
			if(pString[stringPos] != dicArray[searchPos + stringPos])
			{
				stringPos--;
				break;
			}
			stringPos++;
		}
		
		if(stringPos == stringLength)
		{
			return searchIndex + 1;
		}
		
		//add not compared positions to the current position in dictionary
		searchPos += stringLength;
		
	}
	
	return 0;
}
*/

template <const UINT16 DIC_SIZE> 
Vector2D<UINT16> Dictionary<DIC_SIZE>::searchPhrase(BYTE* pString, UINT16 length)
{
	UINT16 searchIndex	= currIndex;
	UINT16 searchPos	= currPos + 1;
	
	UINT16 bestIndex	= 0;
	UINT16 bestLength	= 0;
	
	
	//for each entry
	while(searchIndex--)
	{
		
		//read length of string in current entry
		UINT16 stringLength = dicArray[searchPos++] << 8;
		stringLength 		+=  dicArray[searchPos++];
		
		//if length of searched string and current entry is uneven, then go to next entry
		if((length < stringLength) || (bestLength >= stringLength))
		{
			searchPos += stringLength;
			continue;
		}
		
		//compare searched string and string in entry
		UINT16 stringPos = 0;
		while(stringPos < stringLength)
		{
			if(pString[stringPos] != dicArray[searchPos + stringPos])
			{
				stringPos--;
				break;
			}
			stringPos++;
		}
		
		//add not compared positions to the current position in dictionary
		searchPos += stringLength;
		
		if(stringPos == stringLength)
		{
			bestIndex = searchIndex + 1;
			bestLength = stringLength;
		}

	}
	
	return Vector2D<UINT16>(bestIndex, bestLength);
}


template <const UINT16 DIC_SIZE> 
UINT16 Dictionary<DIC_SIZE>::getString(UINT16 index, BYTE* pString)
{
	UINT16 searchIndex	= currIndex;
	UINT16 searchPos	= currPos + 1;
	
	if(index > currIndex)
	{
		return 0;
	}
	
	while(searchIndex != index)
	{
		//read length of string in current entry
		UINT16 stringLength = dicArray[searchPos++] << 8;
		stringLength 		+=  dicArray[searchPos++];
		
		searchPos += stringLength;
		searchIndex--;
	}
	
	//read length of string in current entry
	UINT16 stringLength = dicArray[searchPos++] << 8;
	stringLength 		+=  dicArray[searchPos++];
	
	for(UINT16 i = 0; i < stringLength; i++)
	{
		pString[i] = dicArray[searchPos++];
	}
	
	return stringLength;
}

#endif
